from math import pi, sin, cos, tan, acos, atan2
from typing import cast
import cadquery as cq


def involute(
    r: float, baseRadius: float, startAngle: float, rev: bool
) -> tuple[float, float]:
    """Calculate points on a involute curve based on radial position as the parameter"""

    alpha = acos(baseRadius / r)
    invAlpha = tan(alpha) - alpha
    if rev:
        return (r * cos(-invAlpha + startAngle), r * sin(-invAlpha + startAngle))
    else:
        return (r * cos(invAlpha + startAngle), r * sin(invAlpha + startAngle))


def involuteProfile(
    module: float,
    teeth: int,
    pressureAngle: float = 20.0,
    rootFillet: float | None = None,
    internal: bool = False,
) -> cq.Wire:
    """Create an involute gear profile at the center of the XY plane"""

    if rootFillet is None:
        rootFillet = 0.38 * module

    pressureAngleRad = pressureAngle * pi / 180
    pitchRadius = module * teeth / 2
    baseRadius = pitchRadius * cos(pressureAngleRad)
    tipRadius = pitchRadius + 1.25 * module if internal else pitchRadius + module
    rootRadius = pitchRadius - module if internal else pitchRadius - 1.25 * module
    # Angular thickness of an involute part of the tooth
    involuteAngle = tan(acos(baseRadius / tipRadius)) - acos(baseRadius / tipRadius)
    pitchInvoluteAngle = tan(pressureAngleRad) - pressureAngleRad
    # Angular thickness of the tip arc (crest)
    # Assuming tooth width on the pitch circle is half of the pitch
    tipAngle = pi / teeth + 2 * (pitchInvoluteAngle - involuteAngle)

    involuteStartRadius = max(baseRadius, rootRadius)
    # Angular offset for connecting a root arc and an involute (when rootRadius < baseRadius)
    (involuteStartX, involuteStartY) = involute(involuteStartRadius, baseRadius, 0, False)
    involuteStartAngle = atan2(involuteStartY, involuteStartX)

    wp = cq.Workplane("XY")
    wp = wp.moveTo(rootRadius * cos(involuteStartAngle), rootRadius * sin(involuteStartAngle))

    for i in range(teeth):
        startAngle = 2 * pi * i / teeth
        endAngle = 2 * pi * (i + 1) / teeth

        # Connection between the root and an involute has two cases
        # https://involutegearsoft.hatenablog.com/entry/2023/09/18/163506
        if rootRadius < baseRadius:
            # Parts near the root are approximated by straight line (unless there are many teeth)
            wp = wp.lineTo(baseRadius * cos(startAngle), baseRadius * sin(startAngle))
        # Draw the outward involute curve
        wp = wp.parametricCurve(
            lambda t: involute(
                (tipRadius - involuteStartRadius) * t + involuteStartRadius,
                baseRadius,
                startAngle,
                False,
            ),
            N=16,
            makeWire=False,
        )
        # Draw the tip of the tooth
        wp = wp.radiusArc(
            (
                tipRadius * cos(startAngle + involuteAngle + tipAngle),
                tipRadius * sin(startAngle + involuteAngle + tipAngle),
            ),
            -tipRadius,
        )
        # Draw the inward involute curve
        wp = wp.parametricCurve(
            lambda t: involute(
                (tipRadius - involuteStartRadius) * (1 - t) + involuteStartRadius,
                baseRadius,
                startAngle + 2 * involuteAngle + tipAngle,
                True,
            ),
            N=16,
            makeWire=False,
        )
        if rootRadius < baseRadius:
            wp = wp.lineTo(
                rootRadius * cos(startAngle + 2 * involuteAngle + tipAngle),
                rootRadius * sin(startAngle + 2 * involuteAngle + tipAngle),
            )
        # Draw the root between teeth
        wp = wp.radiusArc(
            (rootRadius * cos(endAngle + involuteStartAngle), rootRadius * sin(endAngle + involuteStartAngle)), -rootRadius
        )

    wp = wp.close()
    wire = cast(cq.Wire, wp.val())

    # TODO: Fillet for gears with many teeths (currently fillet2D raises error)
    if rootRadius < baseRadius:
        # Fillet only root vertices
        rootVertices = [
            v
            for v in wire.Vertices()
            if v.X * v.X + v.Y * v.Y < (rootRadius * 1.001) ** 2
        ]
        assert len(rootVertices) == 2 * teeth
        wire = wire.fillet2D(
            radius=0.1, vertices=rootVertices
        )  # This has to be fillet2D, not fillet

    return wire



from typing import Literal, cast
from math import pi, sin, cos, tan, atan, sqrt
import cadquery as cq
from OCP.BRepBuilderAPI import BRepBuilderAPI_MakeFace  # type: ignore
from OCP.gp import gp_Sphere  # type: ignore
from .profile import involuteProfile


def makeBevelGear(
    module: float,
    teeth: int,
    counterpartTeeth: int,
    width: float,
    shaftAngle: float = 90.0,
    pressureAngle: float = 20.0,
    rootFillet: float | None = None,
    type: Literal["bevel"] = "bevel",
    flatPartCoeff: float = 0.8,
) -> cq.Solid:
    pitchRadius = module * teeth / 2
    shaftAngleRad = shaftAngle * pi / 180
    coneAngle = atan(
        sin(shaftAngleRad) / ((counterpartTeeth / teeth) + cos(shaftAngleRad))
    )
    coneDistance = pitchRadius / sin(coneAngle)
    coneHeight = pitchRadius / tan(coneAngle)

    profile = involuteProfile(
        module,
        teeth,
        pressureAngle=pressureAngle,
        rootFillet=rootFillet,
    )

    outerSphere = cq.Solid.makeSphere(coneDistance)

    # Project the profile onto a sphere
    projected = cast(
        cq.Wire,
        profile.translate((0, 0, coneHeight)).project(
            cast(cq.Face, outerSphere.faces("not %PLANE")), (0, 0, -1), closest=True
        ),
    )

    # Create spherical face from the projected Wire
    gpSphere = gp_Sphere()
    gpSphere.SetRadius(coneDistance)
    mf = BRepBuilderAPI_MakeFace(gpSphere, projected.wrapped)
    outerFace = cq.Face(mf.Face()).fix()

    # Extrude it to make a bevel gear with specified facewidth
    solid = outerFace.thicken(-width)

    innerConeDistance = coneDistance - width
    innerConeHeight = coneHeight * (innerConeDistance / coneDistance)
    bottomFlatRadius = flatPartCoeff * pitchRadius
    topFlatRadius = bottomFlatRadius * innerConeDistance / coneDistance
    solid = cast(
        cq.Solid,
        solid
        + cq.Solid.makeCylinder(    # Add top spherical part
            topFlatRadius,
            coneHeight - innerConeHeight,
            pnt=(
                0,
                0,
                sqrt(innerConeDistance**2 - topFlatRadius**2),
            )
        )
        - cq.Solid.makeCylinder(    # Remove bottom spherical part
            bottomFlatRadius,
            coneDistance,
            pnt=(
                0,
                0,
                sqrt(coneDistance**2 - bottomFlatRadius**2),
            ),
        ),
    )

    return solid



import argparse
import sys
import cadquery as cq
from .parallel import makeParallelShaftGear
from .bevel import makeBevelGear


def main():
    argParser = argparse.ArgumentParser(prog="geargen")
    argParser.add_argument("-o", "--output", required=True)
    argParser.add_argument("-m", "--module", type=float, required=True)
    argParser.add_argument("-n", "--teeth", type=int, required=True)
    argParser.add_argument("-w", "--width", type=float, required=True)
    argParser.add_argument("--root-fillet", type=float)
    argParser.add_argument("--pressure-angle", type=float, default=20)
    argParser.add_argument("-ha", "--helix_angle", type=float, default=20)
    argParser.add_argument("-t", "--type", default="spur")
    argParser.add_argument("--normal", action="store_true")
    argParser.add_argument("-od", "--outer_diameter", type=float)
    argParser.add_argument("-nc", "--counterpart_teeth", type=int)
    argParser.add_argument("-sa", "--shaft_angle", type=float, default=90)

    args = argParser.parse_args()

    if (
        args.type == "spur"
        or args.type == "helical"
        or args.type == "herringbone"
        or args.type == "internal"
        or args.type == "internal_helical"
        or args.type == "internal_herringbone"
    ):
        if (
            args.type == "internal"
            or args.type == "internal_helical"
            or args.type == "internal_herringbone"
        ) and args.outer_diameter is None:
            print(
                "Outer diameter (-od) is necessary for internal gears", file=sys.stderr
            )
            return -1

        result = makeParallelShaftGear(
            args.module,
            args.teeth,
            args.width,
            pressureAngle=args.pressure_angle,
            rootFillet=args.root_fillet,
            type=args.type,
            helixAngle=args.helix_angle,
            normal=args.normal,
            outerDiameter=args.outer_diameter,
        )
    elif args.type == "bevel":
        result = makeBevelGear(
            args.module,
            args.teeth,
            args.counterpart_teeth,
            args.width,
            shaftAngle=args.shaft_angle,
            pressureAngle=args.pressure_angle,
            rootFillet=args.root_fillet,
            type=args.type,
        )
    else:
        print("Unsupported gear type", file=sys.stderr)
        return -1

    cq.exporters.export(result, args.output)

    return 0



from typing import Literal, cast
from math import pi, cos, tan, atan
import cadquery as cq
from .profile import involuteProfile


def makeParallelShaftGear(
    module: float,
    teeth: int,
    width: float,
    pressureAngle: float = 20.0,
    rootFillet: float | None = None,
    type: Literal[
        "spur",
        "helical",
        "herringbone",
        "internal",
        "internal_helical",
        "internal_herringbone",
    ] = "spur",
    helixAngle: float = 20,
    normal: bool = False,
    outerDiameter: float | None = None,
) -> cq.Solid:
    pressureAngleRad = pressureAngle * pi / 180
    helixAngleRad = pi * helixAngle / 180
    pitchRadius = module * teeth / 2

    if normal and (type == "helical" or type == "herringbone"):
        # Convert normal module and pressure angle into axial ones
        module = module / cos(helixAngleRad)
        pressureAngle = atan(tan(pressureAngleRad) / cos(helixAngleRad)) * 180 / pi

    internal = type in set(["internal", "internal_helical", "internal_herringbone"])

    profile = involuteProfile(
        module,
        teeth,
        pressureAngle=pressureAngle,
        rootFillet=rootFillet,
        internal=internal,
    )

    wp = cq.Workplane(obj=profile).toPending()
    if type == "spur" or type == "internal":
        solid = wp.extrude(width).val()
    elif type == "helical" or type == "internal_helical":
        solid = wp.twistExtrude(
            width, (180 / pi) * width * tan(helixAngleRad) / pitchRadius
        ).val()
    elif type == "herringbone" or type == "internal_herringbone":
        wp = wp.twistExtrude(
            width / 2, (180 / pi) * width * tan(helixAngleRad) / pitchRadius / 2
        )
        solid = wp.mirror(wp.faces("+Z"), union=True).val()
    if internal:
        assert outerDiameter is not None
        solid = (
            cq.Workplane("XY")
            .circle(outerDiameter / 2)
            .extrude(width)
            .cut(cast(cq.Solid, solid))
            .val()
        )
    return cast(cq.Solid, solid)
